unit TestPortAudioDevice;

{* TestPortAudioDevice does some basic tests of the portaudio libs.
 * If all works, it lists all audio input and output devices and their
 * characteristics.
 *}


// This TestPortAudioDevice.pas was slightly changed by BREAKOUTBOX in 2010
// for the use inside a LAZARUS test application.
// For example for the GUI environment writeln() is replaced by MessageBox()
// or moved into Main1.pas for logging into a StringList.
//
// ALSO:  added "procedure LogAddString( var s:string);"
//        to collect the messages for displaying then in a TMemo or so ..
//        and "procedure LogClear();"

// activate this and Messages are written into the TStringList var called "Log":
//{$DEFINE NOLOG}


{* UltraStar Deluxe - Karaoke Game
 *
 * UltraStar Deluxe is the legal property of its developers, whose names
 * are too numerous to list here. Please refer to the COPYRIGHT
 * file distributed with this source distribution.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING. If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 * $URL$
 * $Id$
 *}



{$IFDEF FPC}
  {$mode objfpc}     // .. and activate this for MODE objfpc
  {$H+}
{$ENDIF}


interface

uses
  Interfaces,
  Classes, Dialogs, Forms, SysUtils, math,
  ctypes, crt,
  LConvEncoding,
  PortAudio;


{$IFNDEF NOLOG}
var Log : TStringList;  // write test results into this list ..
{$ENDIF}
var StopStreamTest:boolean;


const
  paDefaultApi = -1;

  ApiPreferenceOrder:
{$IF Defined(MSWINDOWS)}
    // Note1: Portmixer has no mixer support for paASIO and paWASAPI at the moment
    // Note2: Windows Default-API is MME, but DirectSound is faster
    array[0..0] of TPaHostApiTypeId = ( paDirectSound );
{$ELSEIF Defined(DARWIN)}
    array[0..0] of TPaHostApiTypeId = ( paDefaultApi ); // paCoreAudio
{$ELSEIF Defined(UNIX)}
    // Note: Portmixer has no mixer support for JACK at the moment
    array[0..2] of TPaHostApiTypeId = ( paALSA, paJACK, paOSS );
{$ELSE}
    array[0..0] of TPaHostApiTypeId = ( paDefaultApi );
{$IFEND}

  standardSampleRates: array[1..13] of cdouble =
    ( 8000.0,  9600.0,  11025.0,  12000.0,  16000.0,
     22050.0, 24000.0,  32000.0,  44100.0,  48000.0,
     88200.0, 96000.0, 192000.0
    );

  SampleFormat: array[1..8] of culong =
    (paFloat32,      paInt32,          paInt24,      paInt16,      paInt8,
     paUInt8,        paCustomFormat,   paNonInterleaved
    );
  SampleFormatName: array[1..8] of string =
    ('paFloat32',   'paInt32',        'paInt24',    'paInt16',    'paInt8',
     'paUInt8',     'paCustomFormat', 'paNonInterleaved'
    );

{ helper functions }
procedure LogAddString( s:string);
procedure LogClear();

{ test functions }
function  GetPreferredApiIndex: TPaHostApiIndex;
procedure TestInitTerminate;
procedure TestErrorText;
procedure TestVersion;
procedure TestApiInfo;
procedure TestDeviceInfo;
procedure TestFormatInfo;
procedure TestStreams;


implementation

uses main1;

var
  i, j:        integer;
  PaError:     TPaError;
  paApiIndex:  TPaHostApiIndex;
  paApiInfo:   PPaHostApiInfo;
  deviceIndex: TPaDeviceIndex;
  deviceInfo:  PPaDeviceInfo;
  inputParameters:  PPaStreamParameters;
  outputParameters: PPaStreamParameters;
  sampleRate:       cdouble;
  stream:           PPaStream;
  framesPerBuffer:  culong;
  streamFlags:      TPaStreamFlags;
  streamCallback:   PPaStreamCallback;
  callbackStartTime: TDateTime;
  callbackWorks:    boolean;
  userData:         Pointer;


{ ***** Log Functions ******************************************************** }
procedure postMsg( s:string);
begin
  Form1.Memo1.Append( s);
//  PostMessage( Application.MainForm.Handle, WM_USER, v, v);
//  Application.ProcessMessages;
end;

procedure LogAddString( s:string);
begin
  {$IFDEF NOLOG} // if "NOLOG" is defined we instead display a MessageBox !!!
  ShowMessage( s);
  {$ELSE}
  if Log = NIL then
    begin
      log:= TStringList.create;
    end;
  Log.Append( s);
  postMsg( Log.Strings[Log.Count-1]);
  {$ENDIF}
end;

procedure LogClear();
begin
  {$IFNDEF NOLOG}
  if Log <> NIL then Log.Clear;
  {$ENDIF}
end;


{ ***** Test Functions ******************************************************* }
function GetPreferredApiIndex(): TPaHostApiIndex;
var
  i:        integer;
  apiIndex: TPaHostApiIndex;
  apiInfo:  PPaHostApiInfo;
begin
  Result:= -1;
  PaError := Pa_Initialize;

  // select preferred sound-API
  for i:= 0 to High( ApiPreferenceOrder) do
  begin
    if (ApiPreferenceOrder[i] <> paDefaultApi) then
    begin
      // check if API is available
      apiIndex := Pa_HostApiTypeIdToHostApiIndex(ApiPreferenceOrder[i]);
      if (apiIndex >= 0) then
      begin
        // we found an API but we must check if it works
        // (on linux portaudio might detect OSS but does not provide
        // any devices if ALSA is enabled)
        apiInfo := Pa_GetHostApiInfo(apiIndex);
        if (apiInfo^.deviceCount > 0) then
        begin
          Result := apiIndex;
          break;
        end;
      end;
    end;
  end;

  // None of the preferred APIs is available -> use default
  if (Result < 0) then
    begin
      Result := Pa_GetDefaultHostApi();
    end;

  PaError := Pa_Terminate;
end;

{
type
  TAudioSampleFormat = (
    asfU8, asfS8,         // unsigned/signed  8 bits
    asfU16LSB, asfS16LSB, // unsigned/signed 16 bits (endianness: LSB)
    asfU16MSB, asfS16MSB, // unsigned/signed 16 bits (endianness: MSB)
    asfU16, asfS16,       // unsigned/signed 16 bits (endianness: System)
    asfS32,               // signed 32 bits (endianness: System)
    asfFloat,             // float
    asfDouble             // double
  );
  TAudioFormatInfo = ;
  TAudioInputDevice = record
      AudioFormat:     TAudioFormatInfo; // capture format info (e.g. 44.1kHz SInt16 stereo)
      CaptureChannel:  array of TCaptureBuffer; // sound-buffer references used for mono or stereo channel's capture data
  end;

procedure HandleMicrophoneData(Buffer: PByteArray; Size: integer; InputDevice: TAudioInputDevice);
var
  MultiChannelBuffer:      PByteArray;  // buffer handled as array of bytes (offset relative to channel)
  SingleChannelBuffer:     PByteArray;  // temporary buffer for new samples per channel
  SingleChannelBufferSize: integer;
  ChannelIndex:            integer;
  CaptureChannel:          TCaptureBuffer;
  AudioFormat:             TAudioFormatInfo;
  SampleSize:              integer;
  SamplesPerChannel:       integer;
  i:                       integer;
begin
  AudioFormat := InputDevice.AudioFormat;
  SampleSize := AudioSampleSize[AudioFormat.Format];
  SamplesPerChannel := Size div AudioFormat.FrameSize;

  SingleChannelBufferSize := SamplesPerChannel * SampleSize;
  GetMem(SingleChannelBuffer, SingleChannelBufferSize);

  // process channels
  for ChannelIndex := 0 to High(InputDevice.CaptureChannel) do
  begin
    CaptureChannel := InputDevice.CaptureChannel[ChannelIndex];
    // check if a capture buffer was assigned, otherwise there is nothing to do
    if (CaptureChannel <> nil) then
    begin
      // set offset according to channel index
      MultiChannelBuffer := @Buffer[ChannelIndex * SampleSize];
      // separate channel-data from interleaved multi-channel (e.g. stereo) data
      for i := 0 to SamplesPerChannel-1 do
      begin
        Move(MultiChannelBuffer[i*AudioFormat.FrameSize],
             SingleChannelBuffer[i*SampleSize],
             SampleSize);
      end;
      CaptureChannel.ProcessNewBuffer(SingleChannelBuffer, SingleChannelBufferSize);
    end;
  end;

  FreeMem(SingleChannelBuffer);
end;
}

procedure TestInitTerminate();
begin
  LogAddString( '*** Test of Pa_Initialize and Pa_Terminate ***');

  PaError := Pa_Initialize;
  if PaError = paNoError
    then LogAddString( 'Pa_Initialize: No error')
    else LogAddString( 'Pa_Initialize: Error No ' +IntToStr( PaError));

  LogAddString( 'If You use the portaudio.dll from Ultrastar 1.5.2010 22:55:00' +#13#10
                +'then usually an exception occurs when only calling' +#13#10
                +' PA_Initialize + PA_Terminate when closing this Application -' +#13#10
                +' don''t worry .. ;-)');

  PaError := Pa_Terminate;
  if PaError = paNoError
    then LogAddString ('Pa_Terminate:  No error')
    else LogAddString ('Pa_Terminate:  Error No: ' +IntToStr( PaError));
end;

procedure TestErrorText();
begin
  PaError := Pa_Initialize;
  LogAddString( '*** Test of Pa_GetErrorText ***' +#13#10
                +'paNoError (0): ' +Pa_GetErrorText( PaError) +#13#10
                +#13#10
                +'Code      Text' +#13#10
                +'------------------------------------');
  i := paNotInitialized;
  repeat
    LogAddString ( IntToStr(i) +' ' +Pa_GetErrorText(i));
    i := succ(i);
  until SameText( Pa_GetErrorText(i),
                  'Invalid error code') or (i = paNotInitialized + 100);

  LogAddString( IntToStr(i) +' ' +Pa_GetErrorText(i));
  PaError := Pa_Terminate; 
end;

procedure TestVersion();
begin
  LogAddString ('*** Test of Pa_GetVersion and Pa_GetVersionText ***');
  PaError := Pa_Initialize;
  LogAddString( 'Pa_GetVersion:     '
                +IntToStr( Pa_GetVersion) +#13#10
                +'Pa_GetVersionText: ' +Pa_GetVersionText);
  PaError := Pa_Terminate; 
end;

procedure TestApiInfo();
var s:string;
begin
  s:= '*** Test of GetPreferredApiIndex ***'  +#13#10;
  PaError    := Pa_Initialize;
  paApiIndex := GetPreferredApiIndex();
  if (paApiIndex = -1)
    then LogAddString( s +'No working Audio-API found.' +#13#10)
    else LogAddString( s +'working Audio-API found. No: '
                       +IntToStr( paApiIndex) +#13#10);
  PaError := Pa_Terminate; 

  PaError    := Pa_Initialize;
  paApiIndex := GetPreferredApiIndex();
  paApiInfo  := Pa_GetHostApiInfo( paApiIndex);

  LogAddString( '*** Test of Pa_GetHostApiInfo ***' +#13#10
                +'structVersion:' +#9#9
                +IntToStr( paApiInfo^.structVersion)  +#13#10
                +'_type:' +#9#9#9            +IntToStr( paApiInfo^._type) +#13#10
                +'name:' +#9#9#9             +ISO_8859_15ToUTF8( paApiInfo^.name) +#13#10
                +'deviceCount:'+#9#9         +IntToStr( paApiInfo^.deviceCount) +' devices' +#13#10
                +'defaultInputDevice:' +#9   +IntToStr( paApiInfo^.defaultInputDevice) +#13#10
                +'defaultOutputDevice:'+#9   +IntToStr( paApiInfo^.defaultOutputDevice) +#13#10);
  PaError := Pa_Terminate; 

  LogAddString( '*** Test of Pa_HostApiDeviceIndexToDeviceIndex ***');
  PaError    := Pa_Initialize;
  paApiIndex := GetPreferredApiIndex();
  paApiInfo  := Pa_GetHostApiInfo(paApiIndex);
  for i:= 0 to paApiInfo^.deviceCount-1 do
  begin
    deviceIndex := Pa_HostApiDeviceIndexToDeviceIndex(paApiIndex, i);
    LogAddString( 'deviceIndex[' +IntToStr(i) +']: ' +IntToStr( deviceIndex));
  end;
  PaError := Pa_Terminate; 
end;

procedure TestDeviceInfo();
begin
  LogAddString ('*** Test of Pa_GetDeviceCount ***');
  PaError := Pa_Initialize;
  LogAddString ('Pa_GetDeviceCount: ' +IntToStr( Pa_GetDeviceCount) +#13#10);
  PaError := Pa_Terminate;

  LogAddString ('*** Test of Pa_GetDefaultInputDevice ***');
  PaError := Pa_Initialize;
  LogAddString ('Pa_GetDefaultInputDevice: ' +IntToStr( Pa_GetDefaultInputDevice) +#13#10);
  PaError := Pa_Terminate; 

  LogAddString ('*** Test of Pa_GetDefaultOutputDevice');
  PaError := Pa_Initialize;
  LogAddString ('Pa_GetDefaultOutputDevice: ' +IntToStr( Pa_GetDefaultOutputDevice) +#13#10);
  PaError := Pa_Terminate; 

  LogAddString ('*** Test of Pa_GetDeviceInfo ***');
// Note: the fields of deviceInfo can also be used without the '^'.
// deviceInfo.name works as well as deviceInfo^.name
  PaError    := Pa_Initialize;
  paApiIndex := GetPreferredApiIndex();
  paApiInfo  := Pa_GetHostApiInfo(paApiIndex);
  for i:= 0 to paApiInfo^.deviceCount - 1 do
    begin
      deviceIndex := Pa_HostApiDeviceIndexToDeviceIndex(paApiIndex, i);
      deviceInfo  := Pa_GetDeviceInfo(deviceIndex);
      LogAddString( 'deviceInfo of device [' +IntToStr(i) +']:' +#13#10
                    +'name:' +#9#9#9 +ISO_8859_15ToUTF8( deviceInfo^.name) +#13#10
                    +'structVersion:' +#9#9 +IntToStr( deviceInfo^.structVersion) +' (should be 2)' +#13#10
                    +'hostApi:' +#9#9#9 +IntToStr( deviceInfo^.hostApi) +#13#10
                    +'maxInputChannels:' +#9#9 +IntToStr( deviceInfo^.maxInputChannels) +#13#10
                    +'maxOutputChannels:' +#9 +IntToStr( deviceInfo^.maxOutputChannels) +#13#10
                    +'defaultLowInputLatency:' +#9 +FloatToStr( deviceInfo^.defaultLowInputLatency) +#13#10
                    +'defaultLowOutputLatency:' +#9 +FloatToStr( deviceInfo^.defaultLowOutputLatency) +#13#10
                    +'defaultHighInputLatency:' +#9 +FloatToStr( deviceInfo^.defaultHighInputLatency) +#13#10
                    +'defaultHighOutputLatency:' +#9 +FloatToStr( deviceInfo^.defaultHighOutputLatency) +#13#10
                    +'defaultSampleRate:' +#9#9 +FloatToStr( deviceInfo^.defaultSampleRate) +#13#10 );
    end;
  PaError := Pa_Terminate; 
end;

procedure TestFormatInfo();
begin
  LogAddString( '*** Test of Pa_IsFormatSupported ***');
  PaError    := Pa_Initialize;
  paApiIndex := GetPreferredApiIndex();
  paApiInfo  := Pa_GetHostApiInfo(paApiIndex);
  for i:= 0 to paApiInfo^.deviceCount - 1 do
  begin
    deviceIndex := Pa_HostApiDeviceIndexToDeviceIndex(paApiIndex, i);
    deviceInfo  := Pa_GetDeviceInfo(deviceIndex);
    LogAddString ('Device[' +IntToStr(i) +'] ' +ISO_8859_15ToUTF8( deviceInfo^.name) +':');
    New( inputParameters);
    New( outputParameters);

    if deviceInfo^.maxInputChannels > 0 then
    begin
      inputParameters^.device                    := deviceIndex;
      inputParameters^.channelCount              := deviceInfo^.maxInputChannels;
      inputParameters^.sampleFormat              := paInt16;
      inputParameters^.suggestedLatency          := 0;
      inputParameters^.hostApiSpecificStreamInfo := nil;
      outputParameters := nil;
    end
    else
    begin
      inputParameters := nil;
      outputParameters^.device                    := deviceIndex;
      outputParameters^.channelCount              := deviceInfo^.maxOutputChannels;
      outputParameters^.sampleFormat              := paInt16;
      outputParameters^.suggestedLatency          := 0;
      outputParameters^.hostApiSpecificStreamInfo := nil;
    end;

    sampleRate := deviceInfo^.defaultSampleRate;
    PaError    := Pa_IsFormatSupported(inputParameters, outputParameters, sampleRate);
    if PaError = paFormatIsSupported
      then LogAddString( 'Sample rate:' +#9 +FloatToStr( sampleRate) +' : supported')
      else LogAddString( 'Sample rate:' +#9 +FloatToStr( sampleRate) +' : Error: '
                         +Pa_GetErrorText( PaError));

    for j := low(standardSampleRates) to high(standardSampleRates) do
    begin 
      sampleRate := standardSampleRates[j];
      PaError    := Pa_IsFormatSupported(inputParameters, outputParameters, sampleRate);
      if PaError = paFormatIsSupported
        then LogAddString( 'Sample rate:' +#9 +FloatToStr( sampleRate) +' : supported')
        else LogAddString( 'Sample rate:' +#9 +FloatToStr( sampleRate) +' : Error: '
                           +Pa_GetErrorText( PaError));
    end;

    writeln;
    for j := low(SampleFormat) to high(SampleFormat) do
    begin 
      if inputParameters <> nil 
        then inputParameters^.sampleFormat := SampleFormat[j]
        else outputParameters^.sampleFormat := SampleFormat[j];
      PaError := Pa_IsFormatSupported(inputParameters, outputParameters, sampleRate);
      if PaError = paFormatIsSupported
        then LogAddString( 'Sample Format' +#9 +ISO_8859_15ToUTF8( SampleFormatName[j]) +': supported')
        else LogAddString( 'Sample Format' +#9 +ISO_8859_15ToUTF8( SampleFormatName[j]) +': '
                           +Pa_GetErrorText( PaError));
    end;

    Dispose( inputParameters);
    Dispose( outputParameters);
    LogAddString( #13#10);
  end;
  PaError:= Pa_Terminate;
end;


function AudioCallback( {%H-}input: pointer; {%H-}output: pointer; {%H-}frameCount: culong;
                        {%H-}timeInfo: PPaStreamCallbackTimeInfo; {%H-}statusFlags: TPaStreamCallbackFlags;
                        {%H-}inputDevice: pointer): cint; cdecl;
var
  duration : real;
begin
  duration := (Now() - callbackStartTime) * 24 * 3600;
  if (duration < 2.0)
    then result := paContinue
    else
      begin
        callbackWorks := true;
        result := paComplete;
      end;
end;

procedure TestStreams();
var Callback : TPaStreamCallback;  // required in MODE objfpc
    s : string;
    a : dword;
begin
  LogAddString( '*** Test of Pa_OpenStream and Pa_CloseStream ***');
  PaError    := Pa_Initialize;
  paApiIndex := GetPreferredApiIndex();
  paApiInfo  := Pa_GetHostApiInfo( paApiIndex);
  for i:= 0 to paApiInfo^.deviceCount - 1 do
  begin
    if StopStreamTest then
      begin
        postMsg( '*** Test interrupted ***');
        Exit;
      end;
    Application.ProcessMessages;

    deviceIndex := Pa_HostApiDeviceIndexToDeviceIndex(paApiIndex, i);
    deviceInfo  := Pa_GetDeviceInfo(deviceIndex);
    LogAddString( 'Device[' +IntToStr(i) +'] = ' +#9 +UTF8ToAnsi( deviceInfo^.name) +':');
    New( inputParameters);
    New( outputParameters);
    if deviceInfo^.maxInputChannels > 0 then
    begin
      inputParameters^.device                    := deviceIndex;
      inputParameters^.channelCount              := deviceInfo^.maxInputChannels;
      inputParameters^.sampleFormat              := paInt16;
      inputParameters^.suggestedLatency          := deviceInfo^.defaultHighInputLatency;
      inputParameters^.hostApiSpecificStreamInfo := nil;
      outputParameters := nil;
    end
    else
    begin
      inputParameters := nil;
      outputParameters^.device                    := deviceIndex;
      outputParameters^.channelCount              := deviceInfo^.maxOutputChannels;
      outputParameters^.sampleFormat              := paInt16;
      outputParameters^.suggestedLatency          := deviceInfo^.defaultLowOutputLatency;
      outputParameters^.hostApiSpecificStreamInfo := nil;
    end;
      
    sampleRate      := deviceInfo^.defaultSampleRate;
    framesPerBuffer := paFramesPerBufferUnspecified;
    streamFlags     := paNoFlag;

    // in  MODE Delphi ..
    // streamCallback  := @AudioCallback;
    // .. and in MODE objfpc:
    Callback:= @AudioCallback;
    streamCallback  := PPaStreamCallback( Callback);

    userData        := nil;

    Application.ProcessMessages;

    PaError := Pa_OpenStream( stream,
                              inputParameters,
                              outputParameters,
		              sampleRate,
                              framesPerBuffer,
                              streamFlags,
                              streamCallback,
                              userData
		              );
    if (PaError = paNoError) and (stream <> nil)
      then LogAddString( 'Pa_OpenStream:' +#9 +'success')
      else LogAddString( 'Pa_OpenStream:' +#9 +Pa_GetErrorText(PaError));

    Application.ProcessMessages;

    if (PaError = paNoError) and (stream <> nil) then
    begin
      callbackStartTime := Now();

      PaError := Pa_StartStream(stream);
      if (PaError = paNoError)
        then LogAddString ('Pa_StartStream:' +#9 +'success')
        else LogAddString ('Pa_StartStream:' +#9 +Pa_GetErrorText(PaError));

      callbackWorks := false;

      // wait twice the time a successful callback would need for termination
      s:= 'Wait for callback' +#9;
      for a:= 1 to 10 do
        begin
          delay(400);
          Application.ProcessMessages;
        end;

      if (callbackWorks and (Pa_IsStreamStopped(stream) = 0)) then
      begin
	LogAddString( s +'Success: Device works');
	PaError := Pa_StopStream(stream);
	if (PaError = paNoError)
          then LogAddString ('Pa_StopStream:' +#9 +'success')
	  else LogAddString ('Pa_StopStream:' +#9 +Pa_GetErrorText(PaError));
      end
      else
      begin
	LogAddString( s +'Error: Non working device');
        PaError := Pa_AbortStream(stream);
        if (PaError = paNoError)
          then LogAddString ('Pa_AbortStream:' +#9 +'success')
          else LogAddString ('Pa_AbortStream:' +#9 +Pa_GetErrorText(PaError));

      end;
    end;

    Application.ProcessMessages;

    PaError := Pa_CloseStream(stream);
    if PaError = paNoError
      then LogAddString ('Pa_CloseStream:' +#9 +'success')
      else LogAddString( 'Pa_CloseStream:' +#9 +Pa_GetErrorText(PaError));

    Dispose( inputParameters);
    Dispose( outputParameters);
    
    LogAddString( #13#10);
  end;
  PaError := Pa_Terminate; 
  LogAddString( '*** Test executed ***');
  StopStreamTest:= FALSE;
end;


initialization
  //FixMe:  need more information - do we really need this ? :
  // floating point exceptions are raised. Therefore, set the exception mask.
  SetExceptionMask([exZeroDivide, exPrecision]);

  {$IFNDEF NOLOG}
  Log:= NIL;
  {$ENDIF}
  StopStreamTest:= FALSE;

end.

